"""Unit tests for BM25Index configuration validation errors."""

from __future__ import annotations

from unittest.mock import Mock

import pytest
from django.db.backends.base.schema import BaseDatabaseSchemaEditor

from paradedb.indexes import BM25Index
from tests.models import Product


class DummySchemaEditor(BaseDatabaseSchemaEditor):
    """Minimal schema editor for SQL generation in unit tests."""

    def __init__(self) -> None:
        connection = Mock()
        connection.features.uses_case_insensitive_names = False
        super().__init__(connection, collect_sql=False)

    def quote_name(self, name: str) -> str:
        return f'"{name}"'


@pytest.mark.parametrize(
    ("description_config", "error_match"),
    [
        ({"stemmer": "english"}, "no tokenizer"),
        ({"filters": ["lowercase"]}, "no tokenizer"),
        ({"args": [3, 8]}, "no tokenizer"),
        ({"named_args": {"positions": True}}, "no tokenizer"),
        ({"alias": "description_alias"}, "no tokenizer"),
    ],
)
def test_field_config_without_tokenizer_raises_value_error(
    description_config: dict[str, object], error_match: str
) -> None:
    index = BM25Index(
        fields={
            "id": {},
            "description": description_config,
        },
        key_field="id",
        name="product_search_idx",
    )
    with pytest.raises(ValueError, match=error_match):
        index.create_sql(model=Product, schema_editor=DummySchemaEditor())


def test_legacy_options_key_raises_value_error() -> None:
    index = BM25Index(
        fields={
            "id": {},
            "description": {
                "tokenizer": "simple",
                "options": {"remove_long": 20},
            },
        },
        key_field="id",
        name="product_search_idx",
    )
    with pytest.raises(ValueError, match="deprecated 'options'"):
        index.create_sql(model=Product, schema_editor=DummySchemaEditor())


def test_tokenizers_mixed_with_top_level_tokenizer_config_raises_value_error() -> None:
    index = BM25Index(
        fields={
            "id": {},
            "description": {
                "tokenizers": [{"tokenizer": "literal"}],
                "tokenizer": "simple",
            },
        },
        key_field="id",
        name="product_search_idx",
    )
    with pytest.raises(ValueError, match="cannot mix 'tokenizers'"):
        index.create_sql(model=Product, schema_editor=DummySchemaEditor())


@pytest.mark.parametrize(
    ("tokenizers_value", "error_match"),
    [
        ([], "non-empty list"),
        ("not-a-list", "non-empty list"),
        ([123], "must be a dictionary"),
        ([{}], "requires 'tokenizer'"),
    ],
)
def test_invalid_tokenizers_list_shapes_raise_value_error(
    tokenizers_value: object, error_match: str
) -> None:
    index = BM25Index(
        fields={
            "id": {},
            "description": {
                "tokenizers": tokenizers_value,
            },
        },
        key_field="id",
        name="product_search_idx",
    )
    with pytest.raises(ValueError, match=error_match):
        index.create_sql(model=Product, schema_editor=DummySchemaEditor())


def test_json_key_without_tokenizer_raises_value_error() -> None:
    index = BM25Index(
        fields={
            "id": {},
            "metadata": {
                "json_keys": {
                    "color": {},
                }
            },
        },
        key_field="id",
        name="product_search_idx",
    )
    with pytest.raises(ValueError, match="requires an explicit tokenizer"):
        index.create_sql(model=Product, schema_editor=DummySchemaEditor())


def test_native_json_fields_on_non_json_field_raises_value_error() -> None:
    index = BM25Index(
        fields={
            "id": {},
            "description": {
                "json_fields": {"fast": True},
            },
        },
        key_field="id",
        name="product_search_idx",
    )
    with pytest.raises(ValueError, match="is not a JSONField"):
        index.create_sql(model=Product, schema_editor=DummySchemaEditor())


def test_native_json_fields_cannot_mix_with_legacy_json_keys() -> None:
    index = BM25Index(
        fields={
            "id": {},
            "metadata": {
                "json_fields": {"fast": True},
                "json_keys": {"color": {"tokenizer": "literal"}},
            },
        },
        key_field="id",
        name="product_search_idx",
    )
    with pytest.raises(ValueError, match="cannot mix 'json_fields'"):
        index.create_sql(model=Product, schema_editor=DummySchemaEditor())


def test_stemmer_filter_without_stemmer_language_raises_value_error() -> None:
    index = BM25Index(
        fields={
            "id": {},
            "description": {
                "tokenizer": "simple",
                "filters": ["stemmer"],
            },
        },
        key_field="id",
        name="product_search_idx",
    )
    with pytest.raises(ValueError, match="requires a stemmer language"):
        index.create_sql(model=Product, schema_editor=DummySchemaEditor())


def test_pre_rendered_tokenizer_string_with_named_args_is_rendered_verbatim() -> None:
    index = BM25Index(
        fields={
            "id": {},
            "description": {
                "tokenizer": "ngram(3,8)",
                "named_args": {"prefix_only": True},
            },
        },
        key_field="id",
        name="product_search_idx",
    )
    sql = str(index.create_sql(model=Product, schema_editor=DummySchemaEditor()))
    assert "pdb.ngram(3,8)('prefix_only=true')" in sql



"""BM25 index support for Django models."""

from __future__ import annotations

import json
import re
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, cast

from django.db import models
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.backends.ddl_references import Statement
from django.db.models.expressions import Expression
from django.utils.deconstruct import deconstructible

if TYPE_CHECKING:
    ModelField = models.Field[Any, Any]
else:
    ModelField = models.Field


def _quote_term(value: str) -> str:
    escaped = value.replace("'", "''")
    return f"'{escaped}'"


_CONFIG_KEY_RE = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")


def _validate_config_key(key: str, *, context: str) -> str:
    if not isinstance(key, str):
        raise ValueError(f"{context} must be strings.")
    if not _CONFIG_KEY_RE.match(key):
        raise ValueError(
            f"{context} must match {_CONFIG_KEY_RE.pattern!r}; got {key!r}."
        )
    return key


def _render_sql_arg(value: Any) -> str:
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, int | float):
        return str(value)
    if value is None:
        return "null"
    if isinstance(value, str):
        return _quote_term(value)
    raise TypeError(f"Unsupported tokenizer arg type: {type(value).__name__}")


def _render_config_value(value: Any) -> str:
    if isinstance(value, bool):
        return "true" if value else "false"
    if isinstance(value, int | float):
        return str(value)
    if value is None:
        return "null"
    if isinstance(value, str):
        if "=" in value:
            raise ValueError("Tokenizer named arg string values cannot contain '='.")
        if any(ch in value for ch in ("\x00", "\n", "\r")):
            raise ValueError(
                "Tokenizer named arg string values cannot contain control characters."
            )
        return value
    raise TypeError(f"Unsupported tokenizer named arg type: {type(value).__name__}")


def _build_tokenizer_config(
    *,
    tokenizer: str,
    args: list[Any] | None,
    named_args: dict[str, Any] | None,
    filters: list[str] | None,
    stemmer: str | None,
    alias: str | None = None,
) -> str:
    config_parts: dict[str, Any] = {}
    if alias is not None:
        config_parts["alias"] = alias

    if named_args:
        for key, value in named_args.items():
            safe_key = _validate_config_key(key, context="Tokenizer named arg keys")
            config_parts[safe_key] = value

    if filters:
        for name in filters:
            safe_name = _validate_config_key(name, context="Tokenizer filter names")
            if name == "stemmer":
                if stemmer is None:
                    raise ValueError(
                        "Tokenizer filter 'stemmer' requires a stemmer language."
                    )
                config_parts.setdefault("stemmer", stemmer)
            else:
                config_parts.setdefault(safe_name, True)
    elif stemmer:
        config_parts.setdefault("stemmer", stemmer)

    args_sql: list[str] = []
    if args:
        args_sql = [_render_sql_arg(value) for value in args]

    if config_parts:
        rendered_config = ",".join(
            f"{_validate_config_key(key, context='Tokenizer config keys')}={_render_config_value(value)}"
            for key, value in config_parts.items()
        )
        args_sql.append(_quote_term(rendered_config))

    if not args_sql:
        return tokenizer
    return f"{tokenizer}({','.join(args_sql)})"


def _quote_compiled_param(value: Any, schema_editor: BaseDatabaseSchemaEditor) -> str:
    try:
        return schema_editor.quote_value(value)
    except NotImplementedError:
        return _render_sql_arg(value)


def _inline_compiled_params(
    sql: str,
    params: tuple[Any, ...] | list[Any],
    schema_editor: BaseDatabaseSchemaEditor,
) -> str:
    if not params:
        return sql

    sql_parts = sql.split("%s")
    if len(sql_parts) - 1 != len(params):
        raise ValueError(
            "IndexExpression generated unsupported SQL placeholders. "
            "Only simple value placeholders are supported."
        )

    rendered_sql = sql_parts[0]
    for quoted_param, sql_part in zip(
        (_quote_compiled_param(value, schema_editor) for value in params),
        sql_parts[1:],
        strict=True,
    ):
        rendered_sql += quoted_param + sql_part

    return rendered_sql


def _requires_expression_tokenizer(output_field: ModelField | None) -> bool:
    if output_field is None:
        return False
    return isinstance(
        output_field, (models.CharField, models.TextField, models.JSONField)
    )


def _expression_requires_tokenizer(expression: Expression) -> bool:
    output_field = getattr(expression, "output_field", None)
    if output_field is not None:
        return _requires_expression_tokenizer(output_field)

    for source in expression.get_source_expressions():
        if source is not None and _expression_requires_tokenizer(source):
            return True

    return False


def _render_native_json_fields_json(json_fields: dict[str, dict[str, Any]]) -> str:
    return json.dumps(json_fields, separators=(",", ":"), sort_keys=True)


def _validate_native_json_field_config(
    *,
    field: ModelField,
    field_name: str,
    json_fields: Any,
) -> dict[str, Any]:
    if not isinstance(field, models.JSONField):
        raise ValueError(
            f"Field {field_name!r} uses 'json_fields' but is not a JSONField."
        )
    if not isinstance(json_fields, dict):
        raise ValueError(
            f"Field {field_name!r} 'json_fields' must be a dictionary of options."
        )
    return cast(dict[str, Any], json_fields)


def _validate_native_json_field_conflicts(
    *,
    field_name: str,
    json_fields: Any,
    json_keys: Any,
    tokenizers: Any,
    tokenizer: Any,
    args: Any,
    named_args: Any,
    filters: Any,
    stemmer: Any,
    alias: Any,
) -> None:
    if json_fields is None:
        return

    if (
        json_keys is not None
        or tokenizers is not None
        or tokenizer is not None
        or args is not None
        or named_args is not None
        or filters is not None
        or stemmer is not None
        or alias is not None
    ):
        raise ValueError(
            f"Field {field_name!r} cannot mix 'json_fields' with "
            f"'json_keys', 'tokenizers', 'tokenizer', 'args', 'named_args', "
            f"'filters', 'stemmer', or 'alias'."
        )


@deconstructible(path="paradedb.indexes.IndexExpression")
@dataclass
class IndexExpression:
    """Computed expression for BM25 indexing.

    Use this class to index Django expressions (like ``Lower('title')``,
    ``F('rating') + 1``, or ``Concat('first', 'last')``) in a BM25 index.

    For text expressions, specify a tokenizer. For non-text expressions
    (integers, timestamps, etc.), omit the tokenizer to use ``pdb.alias``.

    Args:
        expression: A Django expression to index (e.g., ``Lower('title')``).
        alias: Required. The name used to reference this expression in queries.
        tokenizer: Tokenizer for text expressions (e.g., 'simple', 'unicode_words').
            Omit for non-text expressions to use ``pdb.alias``.
        args: Positional arguments for the tokenizer.
        named_args: Named arguments for the tokenizer configuration.
        filters: Token filters (e.g., ['lowercase', 'stemmer']).
        stemmer: Stemmer language (e.g., 'english').

    Example::

        from django.db.models import F
        from django.db.models.functions import Lower
        from paradedb.indexes import BM25Index, IndexExpression

        BM25Index(
            fields={"id": {}, "description": {}},
            expressions=[
                # Text expression with tokenizer
                IndexExpression(
                    Lower("title"),
                    alias="title_lower",
                    tokenizer="simple",
                ),
                # Non-text expression with pdb.alias
                IndexExpression(
                    F("rating") + 1,
                    alias="rating_plus_one",
                ),
            ],
            key_field="id",
            name="search_idx",
        )
    """

    expression: Expression | str
    alias: str
    tokenizer: str | None = None
    args: list[Any] | None = None
    named_args: dict[str, Any] | None = None
    filters: list[str] | None = None
    stemmer: str | None = None


class BM25Index(models.Index):
    """BM25 index for ParadeDB."""

    suffix = "bm25"

    def __init__(
        self,
        *,
        fields: dict[str, dict[str, Any]],
        key_field: str,
        name: str,
        expressions: list[IndexExpression] | None = None,
        condition: models.Q | None = None,
    ) -> None:
        self.fields_config = fields
        self.key_field = key_field
        self.index_expressions = list(expressions or [])
        super().__init__(name=name, fields=list(fields.keys()), condition=condition)

    def deconstruct(self) -> tuple[str, Any, dict[str, Any]]:
        path, args, kwargs = super().deconstruct()
        kwargs["fields"] = self.fields_config
        kwargs["key_field"] = self.key_field
        kwargs["name"] = self.name
        if self.index_expressions:
            kwargs["expressions"] = self.index_expressions
        return path, args, kwargs

    def create_sql(
        self,
        model: type[models.Model],
        schema_editor: BaseDatabaseSchemaEditor,
        using: str = "",  # noqa: ARG002
        **kwargs: Any,
    ) -> Statement:
        concurrently = kwargs.get("concurrently", False)
        table = schema_editor.quote_name(model._meta.db_table)
        index_name = schema_editor.quote_name(self.name)

        expressions, json_fields = self._build_index_expressions(model, schema_editor)
        expr_sql = ",\n    ".join(expressions)
        storage_params = [f"key_field={_quote_term(self.key_field)}"]
        if json_fields:
            storage_params.append(
                "json_fields="
                + _quote_term(_render_native_json_fields_json(json_fields))
            )

        create_stmt = "CREATE INDEX"
        if concurrently:
            create_stmt += " CONCURRENTLY"
        template = (
            f"{create_stmt} %(name)s ON %(table)s\n"
            "USING bm25 (\n"
            "    %(expressions)s\n"
            ")\n"
            f"WITH ({', '.join(storage_params)})"
        )

        condition_sql = self._get_condition_sql(model, schema_editor)  # type: ignore[attr-defined]
        if condition_sql:
            template += f"\nWHERE {condition_sql}"

        return Statement(
            template,
            name=index_name,
            table=table,
            expressions=expr_sql,
        )

    def _build_index_expressions(
        self, model: type[models.Model], schema_editor: BaseDatabaseSchemaEditor
    ) -> tuple[list[str], dict[str, dict[str, Any]]]:
        expressions: list[str] = []
        json_fields: dict[str, dict[str, Any]] = {}
        for field_name, config in self.fields_config.items():
            field = model._meta.get_field(field_name)
            column_name: str = getattr(field, "column")  # noqa: B009
            column: str = schema_editor.quote_name(column_name)

            json_keys = config.get("json_keys")
            tokenizers = config.get("tokenizers")
            tokenizer = config.get("tokenizer")
            args = config.get("args")
            named_args = self._extract_named_args(config, field_name)
            filters = config.get("filters")
            stemmer = config.get("stemmer")
            alias = config.get("alias")
            native_json_fields = config.get("json_fields")

            _validate_native_json_field_conflicts(
                field_name=field_name,
                json_fields=native_json_fields,
                json_keys=json_keys,
                tokenizers=tokenizers,
                tokenizer=tokenizer,
                args=args,
                named_args=named_args,
                filters=filters,
                stemmer=stemmer,
                alias=alias,
            )
            if native_json_fields is not None:
                expressions.append(column)
                json_fields[field_name] = _validate_native_json_field_config(
                    field=cast(ModelField, field),
                    field_name=field_name,
                    json_fields=native_json_fields,
                )
                continue

            if json_keys:
                expressions.extend(
                    self._build_json_key_expressions(
                        column, field_name, json_keys, schema_editor
                    )
                )
                continue

            if tokenizers is not None:
                if (
                    tokenizer is not None
                    or filters is not None
                    or stemmer is not None
                    or args is not None
                    or named_args is not None
                    or alias is not None
                ):
                    raise ValueError(
                        f"Field {field_name!r} cannot mix 'tokenizers' with "
                        f"'tokenizer', 'args', 'named_args', 'filters', "
                        f"'stemmer', or 'alias'."
                    )
                expressions.extend(
                    self._build_multi_tokenizer_expressions(
                        column=column,
                        field_name=field_name,
                        tokenizers=tokenizers,
                    )
                )
                continue

            if (
                tokenizer is not None
                or filters is not None
                or stemmer is not None
                or args is not None
                or named_args is not None
                or alias is not None
            ):
                if tokenizer is None:
                    raise ValueError(
                        f"Field {field_name!r} specifies tokenizer configuration but "
                        f"no tokenizer. Please set an explicit tokenizer (e.g. "
                        f"'unicode_words', 'simple', 'literal')."
                    )
                tokenizer_sql = _build_tokenizer_config(
                    tokenizer=tokenizer,
                    args=args,
                    named_args=named_args,
                    filters=filters,
                    stemmer=stemmer,
                    alias=alias,
                )
                expressions.append(f"({column}::pdb.{tokenizer_sql})")
            else:
                expressions.append(column)

        # Process IndexExpression entries
        for idx_expr in self.index_expressions:
            expressions.append(
                self._build_computed_expression(idx_expr, model, schema_editor)
            )

        return expressions, json_fields

    def _build_computed_expression(
        self,
        idx_expr: IndexExpression,
        model: type[models.Model],
        schema_editor: BaseDatabaseSchemaEditor,
    ) -> str:
        """Build SQL for a computed expression with pdb.alias or tokenizer cast."""
        raw_expr = idx_expr.expression

        # Handle string field references as F() expressions
        if isinstance(raw_expr, str):
            from django.db.models import F

            expr = cast(Expression, F(raw_expr))
        else:
            expr = raw_expr

        # Compile the expression to SQL
        from django.db.models.sql import Query

        query = Query(model)
        compiler = query.get_compiler(connection=schema_editor.connection)

        # Resolve and compile the expression
        resolved = expr.resolve_expression(query, allow_joins=False, for_save=False)
        if idx_expr.tokenizer is None and _expression_requires_tokenizer(resolved):
            raise ValueError(
                f"IndexExpression with alias {idx_expr.alias!r} resolves to a text or "
                f"JSON value. Specify a tokenizer for text/JSON expressions."
            )
        expr_sql, params = compiler.compile(resolved)
        expr_sql = _inline_compiled_params(expr_sql, params, schema_editor)

        # Build the cast based on whether a tokenizer is specified
        if idx_expr.tokenizer is not None:
            # Text expression with tokenizer
            tokenizer_sql = _build_tokenizer_config(
                tokenizer=idx_expr.tokenizer,
                args=idx_expr.args,
                named_args=idx_expr.named_args,
                filters=idx_expr.filters,
                stemmer=idx_expr.stemmer,
                alias=idx_expr.alias,
            )
            return f"(({expr_sql})::pdb.{tokenizer_sql})"
        else:
            # Non-text expression with pdb.alias
            alias_quoted = _quote_term(idx_expr.alias)
            return f"(({expr_sql})::pdb.alias({alias_quoted}))"

    @staticmethod
    def _extract_named_args(
        config: dict[str, Any], field_name: str
    ) -> dict[str, Any] | None:
        named_args = config.get("named_args")
        if "options" in config:
            raise ValueError(
                f"Field {field_name!r} uses deprecated 'options'. "
                f"Use 'named_args' instead."
            )
        if named_args is not None and not isinstance(named_args, dict):
            raise ValueError(f"Field {field_name!r} 'named_args' must be a dictionary.")
        return named_args

    def _build_multi_tokenizer_expressions(
        self, *, column: str, field_name: str, tokenizers: Any
    ) -> list[str]:
        if not isinstance(tokenizers, list) or not tokenizers:
            raise ValueError(
                f"Field {field_name!r} must define 'tokenizers' as a non-empty list."
            )

        expressions: list[str] = []
        for idx, config in enumerate(tokenizers):
            if not isinstance(config, dict):
                raise ValueError(
                    f"Field {field_name!r} tokenizer entry at position {idx} must be "
                    f"a dictionary."
                )

            tokenizer = config.get("tokenizer")
            if tokenizer is None:
                raise ValueError(
                    f"Field {field_name!r} tokenizer entry at position {idx} "
                    f"requires 'tokenizer'."
                )
            args = config.get("args")
            named_args = self._extract_named_args(config, f"{field_name}[{idx}]")
            filters = config.get("filters")
            stemmer = config.get("stemmer")
            alias = config.get("alias")

            tokenizer_sql = _build_tokenizer_config(
                tokenizer=tokenizer,
                args=args,
                named_args=named_args,
                filters=filters,
                stemmer=stemmer,
                alias=alias,
            )
            expressions.append(f"({column}::pdb.{tokenizer_sql})")

        return expressions

    def _build_json_key_expressions(
        self,
        column: str,
        field_name: str,
        json_keys: dict[str, dict[str, Any]],
        _schema_editor: BaseDatabaseSchemaEditor,
    ) -> list[str]:
        expressions: list[str] = []
        for key, config in json_keys.items():
            tokenizer = config.get("tokenizer")
            if tokenizer is None:
                raise ValueError(
                    f"JSON key {key!r} in field {field_name!r} requires an explicit "
                    f"tokenizer (e.g. 'unicode_words', 'simple', 'literal')."
                )
            args = config.get("args")
            named_args = self._extract_named_args(config, f"{field_name}.{key}")
            filters = config.get("filters")
            stemmer = config.get("stemmer")
            alias = config.get("alias") or f"{field_name}_{key}"
            tokenizer_sql = _build_tokenizer_config(
                tokenizer=tokenizer,
                args=args,
                named_args=named_args,
                filters=filters,
                stemmer=stemmer,
                alias=alias,
            )
            key_literal = _quote_term(key)
            expressions.append(f"(({column}->>{key_literal})::pdb.{tokenizer_sql})")
        return expressions


__all__ = ["BM25Index", "IndexExpression"]



"""Integration tests for ParadeDB JSON faceted queries."""

from __future__ import annotations

import pytest

from paradedb.search import All, ParadeDB
from tests.models import JsonItem, JsonItemNoExpand

pytestmark = [
    pytest.mark.integration,
    pytest.mark.django_db(transaction=True),
    pytest.mark.usefixtures("json_items"),
]


def _keys(facets: dict[str, object]) -> set[object]:
    buckets = facets.get("buckets", [])
    if not isinstance(buckets, list):
        return set()
    return {bucket.get("key") for bucket in buckets if isinstance(bucket, dict)}


def _null_markers(keys: set[object]) -> set[object]:
    return {key for key in keys if str(key).endswith("__PDB_NULL__")}


def _facet(field: str) -> dict[str, object]:
    return JsonItem.objects.filter(id=ParadeDB(All())).facets(field, include_rows=False)


def _facet_no_expand(field: str) -> dict[str, object]:
    return JsonItemNoExpand.objects.filter(id=ParadeDB(All())).facets(
        field, include_rows=False
    )


class TestJsonFacetIntegration:
    """Validate JSON field facet behavior."""

    def test_facets_json_dotted_fields(self) -> None:
        facets = _facet("metadata.color")
        keys = _keys(facets)
        assert {"red", "blue"} <= keys

    def test_facets_json_nested_fields(self) -> None:
        facets = _facet("metadata.deep.nested.value")
        keys = _keys(facets)
        assert {"x", "y"} <= keys
        assert _null_markers(keys)

    def test_facets_json_boolean_field(self) -> None:
        facets = _facet("metadata.flags.featured")
        keys = _keys(facets)
        assert {0, 1} <= keys
        assert _null_markers(keys)

    def test_facets_json_numeric_field(self) -> None:
        facets = _facet("metadata.size")
        keys = _keys(facets)
        assert {3, 5} <= keys

    def test_facets_json_array_field(self) -> None:
        facets = _facet("metadata.tags")
        keys = _keys(facets)
        assert {"a", "b", "c"} <= keys
        assert _null_markers(keys)

    def test_facets_json_numeric_key(self) -> None:
        facets = _facet("metadata.metrics.1")
        keys = _keys(facets)
        assert {"one", "uno"} <= keys
        assert _null_markers(keys)

    def test_facets_json_dot_key_expand(self) -> None:
        facets = _facet("metadata.dot.key")
        keys = _keys(facets)
        assert {"dk", "dk2"} <= keys
        assert _null_markers(keys)

    def test_facets_json_dash_key(self) -> None:
        facets = _facet("metadata.dash-key")
        keys = _keys(facets)
        assert {"d"} <= keys
        assert _null_markers(keys)

    def test_facets_json_space_key(self) -> None:
        facets = _facet("metadata.weird key")
        keys = _keys(facets)
        assert {"space"} <= keys
        assert _null_markers(keys)

    def test_facets_json_array_numbers(self) -> None:
        facets = _facet("metadata.arrnums")
        keys = _keys(facets)
        assert {1, 2, 3} <= keys
        assert _null_markers(keys)

    def test_facets_json_bools_array(self) -> None:
        facets = _facet("metadata.bools")
        keys = _keys(facets)
        assert {0, 1} <= keys
        assert _null_markers(keys)

    def test_facets_json_mixed_types(self) -> None:
        facets = _facet("metadata.mixed")
        keys = _keys(facets)
        assert {"1", 1} <= keys
        assert _null_markers(keys)

    def test_facets_json_null_value(self) -> None:
        facets = _facet("metadata.nullval")
        keys = _keys(facets)
        assert _null_markers(keys)

    def test_facets_json_empty_string(self) -> None:
        facets = _facet("metadata.emptystr")
        keys = _keys(facets)
        assert "" in keys
        assert _null_markers(keys)

    def test_facets_json_float_value(self) -> None:
        facets = _facet("metadata.numfloat")
        keys = _keys(facets)
        assert 2.5 in keys
        assert _null_markers(keys)

    def test_facets_json_object_field(self) -> None:
        facets = _facet("metadata.obj.inner")
        keys = _keys(facets)
        assert {1, 2} <= keys
        assert _null_markers(keys)

    def test_facets_json_object_array_field(self) -> None:
        facets = _facet("metadata.objarray.k")
        keys = _keys(facets)
        assert {"v1", "v2"} <= keys
        assert _null_markers(keys)

    def test_facets_json_object_array_deep_field(self) -> None:
        facets = _facet("metadata.objarray_deep.meta.code")
        keys = _keys(facets)
        assert {"c1", "c2"} <= keys
        assert _null_markers(keys)

    def test_facets_json_nested_array_field(self) -> None:
        facets = _facet("metadata.nested_array")
        keys = _keys(facets)
        assert {"x", "y"} <= keys
        assert _null_markers(keys)

    def test_facets_json_missing_field(self) -> None:
        facets = _facet("metadata.missing")
        keys = _keys(facets)
        assert _null_markers(keys)


@pytest.mark.usefixtures("json_items_no_expand")
class TestJsonFacetNoExpandIntegration:
    """Validate JSON dot-key behavior with expand_dots disabled."""

    def test_facets_json_dot_key_no_expand(self) -> None:
        facets = _facet_no_expand("metadata.dot.key")
        keys = _keys(facets)
        assert _null_markers(keys)



#!/usr/bin/env python
"""Setup script: Create dedicated table for autocomplete demo.

This script creates an autocomplete_items table with product data
from mock_items, optimized for autocomplete with ngram indexes.
"""

import sys
from pathlib import Path

sys.path.insert(0, str(Path(__file__).resolve().parents[1]))
from models import AutocompleteItem


def setup_autocomplete_table() -> int:
    """Create autocomplete_items table with data from mock_items and ngram index."""
    from django.db import connection

    with connection.cursor() as cursor:
        cursor.execute("CREATE EXTENSION IF NOT EXISTS pg_search")

        # Ensure mock_items exists first
        cursor.execute(
            "CALL paradedb.create_bm25_test_table("
            "schema_name => 'public', table_name => 'mock_items')"
        )

        # Drop and recreate table
        print("\nCreating autocomplete_items table...")
        cursor.execute("DROP TABLE IF EXISTS autocomplete_items CASCADE;")
        cursor.execute(
            """
            CREATE TABLE autocomplete_items (
                id SERIAL PRIMARY KEY,
                description TEXT NOT NULL,
                category VARCHAR(100) NOT NULL,
                rating INTEGER NOT NULL,
                in_stock BOOLEAN NOT NULL DEFAULT true,
                created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
            );
            """
        )
        print("  ✓ Table created")

        # Copy data from mock_items
        print("\nCopying data from mock_items...")
        cursor.execute(
            """
            INSERT INTO autocomplete_items (id, description, category, rating, in_stock, created_at)
            SELECT id, description, category, rating, in_stock, created_at
            FROM mock_items;
            """
        )

        cursor.execute("SELECT COUNT(*) FROM autocomplete_items")
        count = cursor.fetchone()[0]
        print(f"  ✓ Copied {count} products from mock_items")

        print("\nCreating autocomplete-optimized BM25 index...")
    with connection.schema_editor(atomic=False) as schema_editor:
        for index in AutocompleteItem._meta.indexes:
            statement = index.create_sql(
                model=AutocompleteItem, schema_editor=schema_editor
            )
            schema_editor.execute(statement)
        print("  ✓ Created BM25 index with:")
        print("    - description (standard tokenizer)")
        print("    - description_ngram (ngram 3-8 for substring matching)")
        print("    - category (literal for exact matching)")

    return count


if __name__ == "__main__":
    print("=" * 60)
    print("Autocomplete Setup - Creating Dedicated Table")
    print("=" * 60)
    count = setup_autocomplete_table()
    print("\n" + "=" * 60)
    print(f"✓ Setup complete! Created autocomplete_items with {count} products")
    print("=" * 60)
    print("\nRun: python examples/autocomplete/autocomplete.py")



"""ParadeDB annotation functions (Score, Snippet, etc.)."""

from __future__ import annotations

from collections.abc import Sequence
from typing import Any, Literal

from django.contrib.postgres.fields import ArrayField
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.models import CharField, F, FloatField, Func, IntegerField, JSONField
from django.db.models.sql.compiler import SQLCompiler

from paradedb.api import (
    FN_AGG,
    FN_SCORE,
    FN_SNIPPET,
    FN_SNIPPET_POSITIONS,
    FN_SNIPPETS,
)

# These diagnostic functions were removed from the pg_search schema but
# django-paradedb keeps the wrappers for backwards compatibility.
FN_INDEXES = "pdb.indexes"
FN_INDEX_SEGMENTS = "pdb.index_segments"
FN_VERIFY_INDEX = "pdb.verify_index"
FN_VERIFY_ALL_INDEXES = "pdb.verify_all_indexes"


def _quote_term(value: str) -> str:
    escaped = value.replace("'", "''")
    return f"'{escaped}'"


def _validate_non_negative_int(name: str, value: int | None) -> None:
    if value is None:
        return
    if isinstance(value, bool) or not isinstance(value, int):
        raise TypeError(f"{name} must be an integer.")
    if value < 0:
        raise ValueError(f"{name} must be zero or positive.")


def _validate_sample_rate(sample_rate: float | None) -> None:
    if sample_rate is None:
        return
    if isinstance(sample_rate, bool) or not isinstance(sample_rate, int | float):
        raise TypeError("sample_rate must be a float between 0.0 and 1.0.")
    if sample_rate < 0 or sample_rate > 1:
        raise ValueError("sample_rate must be between 0.0 and 1.0.")


class Score(Func):
    """BM25 score annotation."""

    function = FN_SCORE
    output_field = FloatField()

    def __init__(self, key_field: str | None = None) -> None:
        expression = F(key_field or "pk")
        super().__init__(expression)


class Snippet(Func):
    """Snippet annotation."""

    function = FN_SNIPPET
    output_field = CharField()

    def __init__(
        self,
        field: str,
        *,
        start_sel: str | None = None,
        stop_sel: str | None = None,
        max_num_chars: int | None = None,
    ) -> None:
        _validate_non_negative_int("max_num_chars", max_num_chars)
        self._formatting = (start_sel, stop_sel, max_num_chars)
        super().__init__(F(field))

    def as_sql(  # type: ignore[override]
        self,
        compiler: SQLCompiler,
        _connection: BaseDatabaseWrapper,
        **_extra_context: Any,
    ) -> tuple[str, list[Any]]:
        field_sql, params = compiler.compile(self.source_expressions[0])
        if params:
            raise ValueError("Snippet does not support parameterized fields.")

        args = [field_sql]
        start_sel, stop_sel, max_num_chars = self._formatting
        if start_sel is not None:
            args.append(_quote_term(start_sel))
        if stop_sel is not None:
            args.append(_quote_term(stop_sel))
        if max_num_chars is not None:
            args.append(str(max_num_chars))

        sql = f"{self.function}({', '.join(args)})"
        return sql, []


class Snippets(Func):
    """Multiple-snippets annotation.

    Wraps ``pdb.snippets(column, ...)`` which returns a text array of all
    matching snippet fragments, with named SQL parameters.

    See: https://docs.paradedb.com/documentation/full-text/highlight#multiple-snippets
    """

    function = FN_SNIPPETS
    output_field = ArrayField(base_field=CharField())

    _VALID_SORT_BY = ("score", "position")

    def __init__(
        self,
        field: str,
        *,
        start_tag: str | None = None,
        end_tag: str | None = None,
        max_num_chars: int | None = None,
        limit: int | None = None,
        offset: int | None = None,
        sort_by: Literal["score", "position"] | None = None,
    ) -> None:
        if sort_by is not None and sort_by not in self._VALID_SORT_BY:
            raise ValueError(
                f"sort_by must be one of {self._VALID_SORT_BY!r}, got {sort_by!r}"
            )
        _validate_non_negative_int("max_num_chars", max_num_chars)
        _validate_non_negative_int("limit", limit)
        _validate_non_negative_int("offset", offset)
        self._start_tag = start_tag
        self._end_tag = end_tag
        self._max_num_chars = max_num_chars
        self._limit = limit
        self._offset = offset
        self._sort_by = sort_by
        super().__init__(F(field))

    def as_sql(  # type: ignore[override]
        self,
        compiler: SQLCompiler,
        _connection: BaseDatabaseWrapper,
        **_extra_context: Any,
    ) -> tuple[str, list[Any]]:
        field_sql, params = compiler.compile(self.source_expressions[0])
        if params:
            raise ValueError("Snippets does not support parameterized fields.")

        parts = [field_sql]
        if self._start_tag is not None:
            parts.append(f"start_tag => {_quote_term(self._start_tag)}")
        if self._end_tag is not None:
            parts.append(f"end_tag => {_quote_term(self._end_tag)}")
        if self._max_num_chars is not None:
            parts.append(f"max_num_chars => {self._max_num_chars}")
        if self._limit is not None:
            parts.append(f'"limit" => {self._limit}')
        if self._offset is not None:
            parts.append(f'"offset" => {self._offset}')
        if self._sort_by is not None:
            parts.append(f"sort_by => {_quote_term(self._sort_by)}")

        sql = f"{self.function}({', '.join(parts)})"
        return sql, []


class SnippetPositions(Func):
    """Byte-offset positions annotation.

    Wraps ``pdb.snippet_positions(column)`` which returns start/end byte
    offset pairs for each matching term.

    See: https://docs.paradedb.com/documentation/full-text/highlight#byte-offsets
    """

    function = FN_SNIPPET_POSITIONS
    output_field = ArrayField(base_field=ArrayField(base_field=IntegerField()))

    def __init__(self, field: str) -> None:
        super().__init__(F(field))

    def as_sql(  # type: ignore[override]
        self,
        compiler: SQLCompiler,
        _connection: BaseDatabaseWrapper,
        **_extra_context: Any,
    ) -> tuple[str, list[Any]]:
        field_sql, params = compiler.compile(self.source_expressions[0])
        if params:
            raise ValueError("SnippetPositions does not support parameterized fields.")

        sql = f"{self.function}({field_sql})"
        return sql, []


class Agg(Func):
    """Aggregate annotation for ParadeDB facets.

    Supports an optional ``filter`` keyword (a ``Q`` object) to emit
    ``FILTER (WHERE ...)`` clauses, enabling conditional aggregations such as::

        MockItem.objects.aggregate(
            electronics=Agg(
                '{"value_count": {"field": "id"}}',
                filter=Q(category=ParadeDB(Term("electronics"))),
            ),
            footwear=Agg(
                '{"value_count": {"field": "id"}}',
                filter=Q(category=ParadeDB(Term("footwear"))),
            ),
        )
    """

    function = FN_AGG
    output_field = JSONField()
    contains_aggregate = True
    window_compatible = True

    def __init__(
        self,
        json_spec: str,
        *,
        exact: bool | None = None,
        filter: Any | None = None,
    ) -> None:
        if exact is not None and not isinstance(exact, bool):
            raise TypeError("Agg exact must be a boolean when provided.")
        self._json_spec = json_spec
        self._exact = exact
        self._filter = filter
        super().__init__()

    _resolved_where: Any | None = None

    def resolve_expression(
        self,
        query: Any = None,
        allow_joins: bool = True,
        reuse: set[str] | None = None,
        summarize: bool = False,
        for_save: bool = False,
    ) -> Agg:
        c = super().resolve_expression(query, allow_joins, reuse, summarize, for_save)
        if self._filter is not None:
            from django.db.models.sql import Query

            if isinstance(query, Query):
                where, _ = query.build_filter(
                    self._filter,
                    can_reuse=reuse,
                    allow_joins=allow_joins,
                    summarize=summarize,
                )
                c._resolved_where = where
        return c

    def as_sql(  # type: ignore[override]
        self,
        compiler: SQLCompiler,
        _connection: BaseDatabaseWrapper,
        **_extra_context: Any,
    ) -> tuple[str, list[Any]]:
        json_literal = _quote_term(self._json_spec)
        if self._exact is False:
            sql = f"{self.function}({json_literal}, false)"
        else:
            sql = f"{self.function}({json_literal})"

        params: list[Any] = []
        resolved_where = getattr(self, "_resolved_where", None)
        if resolved_where is not None:
            filter_sql, filter_params = compiler.compile(resolved_where)
            sql = f"{sql} FILTER (WHERE {filter_sql})"
            params.extend(filter_params)

        return sql, params


def _execute_table_function(
    sql: str,
    params: Sequence[Any],
    *,
    using: str = DEFAULT_DB_ALIAS,
) -> list[dict[str, Any]]:
    with connections[using].cursor() as cursor:
        cursor.execute(sql, params)
        if cursor.description is None:
            return []
        columns = [column[0] for column in cursor.description]
        rows = cursor.fetchall()
    return [dict(zip(columns, row, strict=False)) for row in rows]


def paradedb_indexes(*, using: str = DEFAULT_DB_ALIAS) -> list[dict[str, Any]]:
    """Return metadata for all BM25 indexes from ``pdb.indexes()``."""
    return _execute_table_function(f"SELECT * FROM {FN_INDEXES}()", (), using=using)


def paradedb_index_segments(
    index: str, *, using: str = DEFAULT_DB_ALIAS
) -> list[dict[str, Any]]:
    """Return segment metadata for a BM25 index from ``pdb.index_segments()``."""
    return _execute_table_function(
        f"SELECT * FROM {FN_INDEX_SEGMENTS}(%s::regclass)", (index,), using=using
    )


def paradedb_verify_index(
    index: str,
    *,
    heapallindexed: bool = False,
    sample_rate: float | None = None,
    report_progress: bool = False,
    verbose: bool = False,
    on_error_stop: bool = False,
    segment_ids: Sequence[int] | None = None,
    using: str = DEFAULT_DB_ALIAS,
) -> list[dict[str, Any]]:
    """Run ``pdb.verify_index()`` for one BM25 index."""
    _validate_sample_rate(sample_rate)
    sql = [f"SELECT * FROM {FN_VERIFY_INDEX}(%s::regclass"]
    params: list[Any] = [index]
    if heapallindexed:
        sql.append(", heapallindexed => %s::boolean")
        params.append(heapallindexed)
    if sample_rate is not None:
        sql.append(", sample_rate => %s::double precision")
        params.append(sample_rate)
    if report_progress:
        sql.append(", report_progress => %s::boolean")
        params.append(report_progress)
    if verbose:
        sql.append(", verbose => %s::boolean")
        params.append(verbose)
    if on_error_stop:
        sql.append(", on_error_stop => %s::boolean")
        params.append(on_error_stop)
    if segment_ids is not None:
        sql.append(", segment_ids => %s::int[]")
        params.append(list(segment_ids))
    sql.append(")")
    return _execute_table_function("".join(sql), params, using=using)


def paradedb_verify_all_indexes(
    *,
    schema_pattern: str | None = None,
    index_pattern: str | None = None,
    heapallindexed: bool = False,
    sample_rate: float | None = None,
    report_progress: bool = False,
    on_error_stop: bool = False,
    using: str = DEFAULT_DB_ALIAS,
) -> list[dict[str, Any]]:
    """Run ``pdb.verify_all_indexes()`` across BM25 indexes."""
    _validate_sample_rate(sample_rate)
    sql = [f"SELECT * FROM {FN_VERIFY_ALL_INDEXES}("]
    params: list[Any] = []
    named_params: list[tuple[str, str, Any]] = []
    if schema_pattern is not None:
        named_params.append(("schema_pattern", "text", schema_pattern))
    if index_pattern is not None:
        named_params.append(("index_pattern", "text", index_pattern))
    if heapallindexed:
        named_params.append(("heapallindexed", "boolean", heapallindexed))
    if sample_rate is not None:
        named_params.append(("sample_rate", "double precision", sample_rate))
    if report_progress:
        named_params.append(("report_progress", "boolean", report_progress))
    if on_error_stop:
        named_params.append(("on_error_stop", "boolean", on_error_stop))

    if named_params:
        sql.append(
            ", ".join(
                f"{parameter_name} => %s::{parameter_type}"
                for parameter_name, parameter_type, _parameter_value in named_params
            )
        )
        params.extend(
            parameter_value
            for _parameter_name, _parameter_type, parameter_value in named_params
        )

    sql.append(")")
    return _execute_table_function("".join(sql), params, using=using)


__all__ = [
    "Agg",
    "Score",
    "Snippet",
    "SnippetPositions",
    "Snippets",
    "paradedb_index_segments",
    "paradedb_indexes",
    "paradedb_verify_all_indexes",
    "paradedb_verify_index",
]



"""Unit tests for api.json loading and validation helpers."""

from __future__ import annotations

import pytest

from paradedb.api import _validate_api_payload


def test_validate_api_payload_accepts_expected_shape() -> None:
    payload = {
        "operators": {"OP_SEARCH": "@@@"},
        "functions": {"FN_ALL": "pdb.all"},
        "types": {"PDB_TYPE_QUERY": "pdb.query"},
    }
    validated = _validate_api_payload(payload)
    assert validated == payload


def test_validate_api_payload_rejects_missing_sections() -> None:
    payload = {
        "operators": {"OP_SEARCH": "@@@"},
        "functions": {"FN_ALL": "pdb.all"},
    }
    with pytest.raises(ValueError, match="section 'types' must be an object"):
        _validate_api_payload(payload)


def test_validate_api_payload_rejects_non_string_values() -> None:
    payload = {
        "operators": {"OP_SEARCH": "@@@"},
        "functions": {"FN_ALL": 1},
        "types": {"PDB_TYPE_QUERY": "pdb.query"},
    }
    with pytest.raises(ValueError, match="must map to a string"):
        _validate_api_payload(payload)



"""Shared model definitions for autocomplete examples."""

from common import configure_django
from django.db import models

from paradedb.indexes import BM25Index
from paradedb.queryset import ParadeDBManager

configure_django()


class AutocompleteItem(models.Model):
    id = models.IntegerField(primary_key=True)
    description = models.TextField()
    category = models.CharField(max_length=100)
    rating = models.IntegerField()
    in_stock = models.BooleanField()
    created_at = models.DateTimeField()

    objects = ParadeDBManager()

    class Meta:
        app_label = "examples"
        managed = False
        db_table = "autocomplete_items"
        indexes = (
            BM25Index(
                fields={
                    "id": {},
                    "description": {
                        "tokenizers": [
                            {"tokenizer": "unicode_words"},
                            {
                                "tokenizer": "ngram",
                                "args": [3, 8],
                                "alias": "description_ngram",
                            },
                        ]
                    },
                    "category": {"tokenizer": "literal", "alias": "category"},
                },
                key_field="id",
                name="autocomplete_items_idx",
            ),
        )

    def __str__(self) -> str:
        return self.description



"""Pytest configuration shared across unit and integration tests."""

from __future__ import annotations

import django
from django.conf import settings

from tests._db_config import database_settings


def pytest_configure(config: object) -> None:
    """Ensure Django is initialized and register custom markers."""
    config.addinivalue_line(
        "markers",
        "integration: marks tests that require a ParadeDB Postgres instance",
    )

    if not settings.configured:
        settings.configure(
            INSTALLED_APPS=["django.contrib.contenttypes", "tests"],
            DATABASES={"default": database_settings()},
            DEFAULT_AUTO_FIELD="django.db.models.BigAutoField",
            SECRET_KEY="tests-secret-key",
            MIGRATION_MODULES={"tests": None},
        )
    elif "postgresql" not in settings.DATABASES.get("default", {}).get("ENGINE", ""):
        settings.DATABASES["default"] = database_settings()

    django.setup()



"""Edge case tests for ParadeDB Django integration.

Tests for special characters, validation, boundary conditions, and unusual inputs.
These are unit tests that don't require a database.
"""

from __future__ import annotations

import pytest

from paradedb.search import (
    Boost,
    Const,
    Match,
    MoreLikeThis,
    ParadeDB,
    Parse,
    Phrase,
    PhrasePrefix,
    Proximity,
    ProximityNode,
    ProxRegex,
    Regex,
    RegexPhrase,
    Term,
)
from tests.models import Product


class TestSpecialCharacterEscaping:
    """Test SQL injection prevention and special character handling."""

    def test_single_quote_in_search_term(self) -> None:
        """Single quotes are escaped to prevent SQL injection."""
        queryset = Product.objects.filter(
            description=ParadeDB(Match("it's", operator="AND"))
        )
        sql = str(queryset.query)
        assert "it''s" in sql

    def test_double_single_quotes(self) -> None:
        """Multiple single quotes are all escaped."""
        queryset = Product.objects.filter(
            description=ParadeDB(Match("don''t", operator="AND"))
        )
        sql = str(queryset.query)
        assert "don''''t" in sql

    def test_backslash_in_search_term(self) -> None:
        """Backslashes are preserved in search terms."""
        queryset = Product.objects.filter(
            description=ParadeDB(Match("path\\to\\file", operator="AND"))
        )
        sql = str(queryset.query)
        assert "path\\to\\file" in sql

    def test_unicode_characters(self) -> None:
        """Unicode characters work in search terms."""
        queryset = Product.objects.filter(
            description=ParadeDB(Match("日本語", operator="AND"))
        )
        sql = str(queryset.query)
        assert "日本語" in sql

    def test_emoji_in_search(self) -> None:
        """Emoji characters work in search terms."""
        queryset = Product.objects.filter(
            description=ParadeDB(Match("👟 shoes", operator="AND"))
        )
        sql = str(queryset.query)
        assert "👟 shoes" in sql

    def test_special_sql_keywords(self) -> None:
        """SQL keywords in search terms are quoted safely."""
        queryset = Product.objects.filter(
            description=ParadeDB(Match("SELECT * FROM", operator="AND"))
        )
        sql = str(queryset.query)
        assert "'SELECT * FROM'" in sql

    def test_phrase_with_quotes(self) -> None:
        """Phrase containing quotes is escaped."""
        queryset = Product.objects.filter(
            description=ParadeDB(Phrase('it\'s a "test"'))
        )
        sql = str(queryset.query)
        assert "it''s" in sql

    def test_regex_special_chars_preserved(self) -> None:
        """Regex special characters are preserved."""
        queryset = Product.objects.filter(
            description=ParadeDB(Regex("test.*[a-z]+\\d{2,3}"))
        )
        sql = str(queryset.query)
        assert "test.*[a-z]+\\d{2,3}" in sql


class TestParadeDBValidation:
    """Test input validation for ParadeDB wrapper."""

    def test_paradedb_invalid_tokenizer_deferred_to_database(self) -> None:
        """Tokenizer names are quoted in SQL; validity is deferred to database execution."""
        queryset = Product.objects.filter(
            description=ParadeDB(
                Match("shoes", operator="AND", tokenizer="bad-tokenizer;")
            )
        )
        assert '::pdb."bad-tokenizer;"' in str(queryset.query)


class TestPhraseValidation:
    """Test Phrase dataclass validation."""

    def test_phrase_slop_zero_is_valid(self) -> None:
        """Slop of 0 is valid."""
        phrase = Phrase("test", slop=0)
        assert phrase.slop == 0

    def test_phrase_negative_slop_raises(self) -> None:
        """Negative slop raises ValueError."""
        with pytest.raises(ValueError, match="zero or positive"):
            Phrase("test", slop=-1)

    def test_phrase_large_slop(self) -> None:
        """Large slop values work."""
        phrase = Phrase("test", slop=100)
        assert phrase.slop == 100

    def test_phrase_must_include_at_least_one_term(self) -> None:
        with pytest.raises(ValueError, match="Phrase requires at least one term"):
            Phrase()

    def test_phrase_text_must_be_string(self) -> None:
        with pytest.raises(TypeError, match="Phrase term must be a string"):
            Phrase(123)  # type: ignore[arg-type]

    def test_phrase_slop_must_be_integer(self) -> None:
        with pytest.raises(TypeError, match="Phrase slop must be an integer"):
            Phrase("test", slop=True)  # type: ignore[arg-type]

    def test_phrase_tokenizer_must_be_string(self) -> None:
        with pytest.raises(TypeError, match="Phrase tokenizer must be a string"):
            Phrase("test", tokenizer=1)  # type: ignore[arg-type]

    def test_phrase_invalid_tokenizer_deferred_to_database(self) -> None:
        """Phrase tokenizer names are quoted in SQL; validity is deferred to database execution."""
        queryset = Product.objects.filter(
            description=ParadeDB(Phrase("test", tokenizer="bad tokenizer"))
        )
        assert '::pdb."bad tokenizer"' in str(queryset.query)


class TestDistanceValidation:
    """Test distance validation on Match and Term."""

    def test_match_default_distance(self) -> None:
        match = Match("test", operator="AND")
        assert match.distance is None

    def test_match_distance_zero_is_valid(self) -> None:
        match = Match("test", operator="AND", distance=0)
        assert match.distance == 0

    def test_match_negative_distance_raises(self) -> None:
        with pytest.raises(ValueError, match="between 0 and 2, inclusive"):
            Match("test", operator="AND", distance=-1)

    def test_match_large_distance_raises(self) -> None:
        with pytest.raises(ValueError, match="between 0 and 2, inclusive"):
            Match("test", operator="AND", distance=10)

    def test_match_distance_must_be_integer(self) -> None:
        with pytest.raises(TypeError, match="Distance must be an integer"):
            Match("test", operator="AND", distance=True)  # type: ignore[arg-type]

    def test_term_distance_validation(self) -> None:
        term = Term("test", distance=1)
        assert term.distance == 1

    def test_term_distance_must_be_integer(self) -> None:
        with pytest.raises(TypeError, match="Distance must be an integer"):
            Term("test", distance=True)  # type: ignore[arg-type]

    def test_match_tokenizer_and_fuzzy_options_rejected(self) -> None:
        with pytest.raises(
            ValueError, match="Match tokenizer cannot be combined with fuzzy options"
        ):
            Match("test", operator="AND", tokenizer="whitespace", distance=1)

    def test_match_multi_term_fuzzy_with_boost_rejected(self) -> None:
        with pytest.raises(
            ValueError, match="Multi-term fuzzy Match does not support boost or const"
        ):
            Match("a", "b", operator="OR", distance=1, boost=2.0)

    def test_match_multi_term_fuzzy_with_const_rejected(self) -> None:
        with pytest.raises(
            ValueError, match="Multi-term fuzzy Match does not support boost or const"
        ):
            Match("a", "b", operator="OR", distance=1, const=1.0)

    def test_term_non_string_rejected(self) -> None:
        with pytest.raises(TypeError, match="Term text must be a string"):
            Term(123)  # type: ignore[arg-type]

    def test_match_non_string_terms_rejected(self) -> None:
        with pytest.raises(TypeError, match="Match terms must be strings"):
            Match("ok", 123, operator="AND")  # type: ignore[arg-type]

    def test_match_non_string_tokenizer_rejected(self) -> None:
        with pytest.raises(TypeError, match="Match tokenizer must be a string"):
            Match("ok", operator="AND", tokenizer=123)  # type: ignore[arg-type]

    def test_match_prefix_must_be_boolean(self) -> None:
        with pytest.raises(TypeError, match="Match prefix must be a boolean"):
            Match("ok", operator="AND", prefix=1)  # type: ignore[arg-type]

    def test_term_prefix_must_be_boolean(self) -> None:
        with pytest.raises(TypeError, match="Term prefix must be a boolean"):
            Term("ok", prefix=1)  # type: ignore[arg-type]


class TestExpressionValidation:
    """Validation coverage for expression dataclasses."""

    def test_proximity_boost_and_const_wrap_query_node(self) -> None:
        proximity = Proximity("a").within(1, "b")
        boosted = proximity.boost(1.0)
        constant = proximity.const(1.0)
        assert boosted.node == proximity
        assert boosted.relevance_modifier == Boost(1.0)
        assert constant.node == proximity
        assert constant.relevance_modifier == Const(1.0)

    def test_proximity_start_must_be_string_or_proxregex(self) -> None:
        with pytest.raises(
            TypeError, match="Proximity term must be strings or ProxRegex instances"
        ):
            Proximity(1)  # type: ignore[arg-type]

    def test_parse_boost_and_const_deferred_to_database(self) -> None:
        parsed = Parse("query", boost=1.0, const=1.0)
        assert parsed.boost == 1.0
        assert parsed.const == 1.0

    def test_parse_query_must_be_string(self) -> None:
        with pytest.raises(TypeError, match="Parse query must be a string"):
            Parse(123)  # type: ignore[arg-type]

    def test_parse_lenient_must_be_boolean(self) -> None:
        with pytest.raises(TypeError, match="Parse lenient must be a boolean"):
            Parse("query", lenient="yes")  # type: ignore[arg-type]

    def test_regex_phrase_requires_terms(self) -> None:
        with pytest.raises(ValueError, match="requires at least one regex term"):
            RegexPhrase()

    def test_regex_phrase_negative_slop_raises(self) -> None:
        with pytest.raises(ValueError, match="slop must be zero or positive"):
            RegexPhrase("a.*", slop=-1)

    def test_regex_phrase_terms_must_be_strings(self) -> None:
        with pytest.raises(TypeError, match="RegexPhrase regex must be a string"):
            RegexPhrase(123)  # type: ignore[arg-type]

    def test_regex_phrase_max_expansions_must_be_integer(self) -> None:
        with pytest.raises(
            TypeError, match="RegexPhrase max_expansions must be an integer"
        ):
            RegexPhrase("a.*", max_expansions=True)  # type: ignore[arg-type]

    def test_regex_phrase_boost_and_const_deferred_to_database(self) -> None:
        regex_phrase = RegexPhrase("a.*", boost=1.0, const=1.0)
        assert regex_phrase.boost == 1.0
        assert regex_phrase.const == 1.0

    def test_phrase_prefix_terms_must_be_strings(self) -> None:
        with pytest.raises(TypeError, match="PhrasePrefix phrase must be a string"):
            PhrasePrefix("ok", 1)  # type: ignore[arg-type]

    def test_phrase_prefix_max_expansion_must_be_integer(self) -> None:
        with pytest.raises(
            TypeError, match="PhrasePrefix max_expansion must be an integer"
        ):
            PhrasePrefix("ok", "term", max_expansion=True)  # type: ignore[arg-type]

    def test_regex_pattern_must_be_string(self) -> None:
        with pytest.raises(TypeError, match="Regex pattern must be a string"):
            Regex(123)  # type: ignore[arg-type]

    def test_proximity_negative_distance_raises(self) -> None:
        with pytest.raises(
            ValueError, match="Proximity distance must be zero or positive"
        ):
            Proximity("left").within(-1, ProxRegex("right.*"))

    def test_prox_regex_negative_max_expansions_in_chain_raises(self) -> None:
        with pytest.raises(ValueError, match="max_expansions must be zero or positive"):
            Proximity("left").within(1, ProxRegex("right.*", max_expansions=-1))

    def test_proximity_with_prox_regex_boost_and_const_wrap_query_node(self) -> None:
        proximity = Proximity("left").within(1, ProxRegex("right.*"))
        assert proximity.boost(1.0).relevance_modifier == Boost(1.0)
        assert proximity.const(1.0).relevance_modifier == Const(1.0)

    def test_proximity_allows_empty_start_term_list(self) -> None:
        proximity = Proximity([])
        assert proximity.term == []

    def test_proximity_then_allows_empty_term_list(self) -> None:
        proximity = Proximity("left").within(1, [])
        assert proximity.right == []

    def test_proximity_then_negative_distance_raises(self) -> None:
        with pytest.raises(
            ValueError, match="Proximity distance must be zero or positive"
        ):
            Proximity("left").within(-1, "right")

    def test_proximity_then_preserves_non_boolean_ordered_value(self) -> None:
        proximity = Proximity("left").within(1, "right", ordered=1)  # type: ignore[arg-type]
        assert proximity.ordered == 1

    def test_proximity_boost_and_const_with_step_wrap_query_node(self) -> None:
        proximity = Proximity("left").within(1, "right")
        assert proximity.boost(1.0).node == proximity
        assert proximity.const(1.0).node == proximity

    def test_proximity_accepts_prox_regex_items_in_start(self) -> None:
        proximity = Proximity(["chicken", ProxRegex("r..s")]).within(1, "delicious")
        assert isinstance(proximity.left, list)
        assert len(proximity.left) == 2
        assert proximity.left[0] == "chicken"
        assert isinstance(proximity.left[1], ProxRegex)
        assert proximity.left[1].pattern == "r..s"

    def test_proximity_accepts_term_lists_in_step(self) -> None:
        proximity = Proximity("chicken").within(1, ["delicious", ProxRegex("cris.*")])
        assert isinstance(proximity.right, list)
        assert len(proximity.right) == 2
        assert proximity.right[0] == "delicious"
        assert isinstance(proximity.right[1], ProxRegex)
        assert proximity.right[1].pattern == "cris.*"

    def test_prox_regex_negative_max_expansions_raises(self) -> None:
        with pytest.raises(ValueError, match="max_expansions must be zero or positive"):
            ProxRegex("pattern", max_expansions=-1)

    def test_prox_regex_max_expansions_must_be_integer(self) -> None:
        with pytest.raises(TypeError, match="max_expansions must be an integer"):
            ProxRegex("pattern", max_expansions=1.5)  # type: ignore[arg-type]

    def test_prox_regex_pattern_must_be_string(self) -> None:
        with pytest.raises(TypeError, match="pattern must be a string"):
            ProxRegex(123)  # type: ignore[arg-type]

    def test_prox_regex_defaults(self) -> None:
        prox_regex = ProxRegex("pattern")
        assert prox_regex.pattern == "pattern"
        assert prox_regex.max_expansions is None

    def test_proximity_start_must_be_strings_or_proxregex(self) -> None:
        with pytest.raises(
            TypeError,
            match="term must be strings or ProxRegex instances",
        ):
            Proximity(123)  # type: ignore[arg-type]

    def test_proximity_step_preserves_unvalidated_term_value(self) -> None:
        proximity = Proximity("left").within(1, 123)  # type: ignore[arg-type]
        assert proximity.right == 123

    def test_proximity_chain_builder_preserves_ordering(self) -> None:
        chain = Proximity("left").within(
            1,
            ["middle", ProxRegex("r.*")],
            ordered=True,
        )
        assert chain.ordered is True

    def test_then_accepts_explicit_grouped_proximity_child(self) -> None:
        grouped = ProximityNode(
            1,
            False,
            "middle",
            ProximityNode(2, True, "right", "tail"),
        )
        chain = Proximity("left").within(3, grouped)
        assert chain.right == grouped
        assert chain.left == "left"
        assert chain.distance == 3
        assert chain.ordered is False

    def test_proximity_node_accepts_nested_proximity_children(self) -> None:
        child = ProximityNode(2, True, "right", "tail")
        node = ProximityNode(1, False, "left", child)
        assert node.left == "left"
        assert node.right == child
        assert node.distance == 1

    def test_root_proximity_rejects_step_only_options(self) -> None:
        with pytest.raises(TypeError):
            Proximity("left", ordered=True)  # type: ignore[call-arg]

    def test_proximity_step_requires_distance(self) -> None:
        with pytest.raises(TypeError):
            ProximityNode("left")  # type: ignore[call-arg]

    def test_top_level_proximity_step_renders_directly(self) -> None:
        queryset = Product.objects.filter(
            description=ParadeDB(ProximityNode(1, False, "right", "tail"))
        )
        assert (
            str(queryset.query)
            == 'SELECT "tests_product"."id", "tests_product"."description", "tests_product"."category", "tests_product"."rating", "tests_product"."in_stock", "tests_product"."price", "tests_product"."created_at", "tests_product"."metadata" FROM "tests_product" WHERE "tests_product"."description" @@@ (\'right\' ## 1 ## \'tail\')'
        )


class TestMoreLikeThisValidation:
    """Test MoreLikeThis validation."""

    def test_mlt_requires_one_input(self) -> None:
        """MLT with no inputs raises ValueError."""
        with pytest.raises(ValueError, match="exactly one input"):
            MoreLikeThis()

    def test_mlt_multiple_inputs_raises(self) -> None:
        """MLT with multiple inputs raises ValueError."""
        with pytest.raises(ValueError, match="exactly one input"):
            MoreLikeThis(id=1, document={"description": "test"})

    def test_mlt_empty_ids_raises(self) -> None:
        """MLT with empty ids raises ValueError."""
        with pytest.raises(ValueError, match="cannot be empty"):
            MoreLikeThis(ids=[])

    def test_mlt_single_id(self) -> None:
        """MLT with single id works."""
        mlt = MoreLikeThis(id=1)
        assert mlt.id == 1

    def test_mlt_id_must_be_integer(self) -> None:
        with pytest.raises(TypeError, match="id must be an integer"):
            MoreLikeThis(id="1")  # type: ignore[arg-type]

        with pytest.raises(TypeError, match="id must be an integer"):
            MoreLikeThis(id=True)  # type: ignore[arg-type]

    def test_mlt_ids_list(self) -> None:
        """MLT with ids list works."""
        mlt = MoreLikeThis(ids=[1, 2, 3])
        assert mlt.ids == [1, 2, 3]

    def test_mlt_ids_must_contain_integers(self) -> None:
        with pytest.raises(TypeError, match="ids must contain integers"):
            MoreLikeThis(ids=[1, "2"])  # type: ignore[list-item]

        with pytest.raises(TypeError, match="ids must contain integers"):
            MoreLikeThis(ids=[1, True])  # type: ignore[list-item]

    def test_mlt_document_dict(self) -> None:
        """MLT with document dict works."""
        mlt = MoreLikeThis(document={"description": "running shoes"})
        assert mlt.document == '{"description": "running shoes"}'

    def test_mlt_document_string_must_be_valid_json_object(self) -> None:
        with pytest.raises(ValueError, match="must decode to an object"):
            MoreLikeThis(document='["not", "an", "object"]')

        with pytest.raises(ValueError, match="must be valid JSON"):
            MoreLikeThis(document="{bad json}")

    def test_mlt_document_with_fields_raises(self) -> None:
        """MLT with document and fields raises ValueError."""
        with pytest.raises(ValueError, match="fields are only valid"):
            MoreLikeThis(
                document={"description": "running shoes"}, fields=["description"]
            )

    def test_mlt_fields_must_be_non_empty_strings(self) -> None:
        with pytest.raises(ValueError, match="fields cannot be empty"):
            MoreLikeThis(id=1, fields=[])

        with pytest.raises(TypeError, match="fields must contain strings"):
            MoreLikeThis(id=1, fields=[1])  # type: ignore[list-item]

        with pytest.raises(ValueError, match="fields cannot contain empty names"):
            MoreLikeThis(id=1, fields=[""])

    def test_mlt_with_all_options(self) -> None:
        """MLT with all tuning options works."""
        mlt = MoreLikeThis(
            id=1,
            fields=["description"],
            min_term_freq=2,
            max_query_terms=10,
            min_doc_freq=1,
            max_term_freq=100,
            max_doc_freq=1000,
            min_word_length=3,
            max_word_length=15,
            stopwords=["the", "a"],
        )
        assert mlt.min_term_freq == 2
        assert mlt.max_query_terms == 10
        assert mlt.min_word_length == 3
        assert mlt.max_word_length == 15
        assert mlt.stopwords == ["the", "a"]

    def test_mlt_stopwords_empty_list(self) -> None:
        """MLT with empty stopwords list works."""
        mlt = MoreLikeThis(id=1, stopwords=[])
        assert mlt.stopwords == []

    def test_mlt_stopwords_tuple(self) -> None:
        """MLT with stopwords as tuple converts to list."""
        mlt = MoreLikeThis(id=1, stopwords=("the", "a", "an"))
        assert mlt.stopwords == ["the", "a", "an"]

    def test_mlt_stopwords_must_be_strings(self) -> None:
        with pytest.raises(TypeError, match="stopwords must contain strings"):
            MoreLikeThis(id=1, stopwords=["the", 1])  # type: ignore[list-item]

    def test_mlt_word_length_validation(self) -> None:
        """MLT word length parameters accept integers."""
        mlt = MoreLikeThis(
            id=1,
            min_word_length=2,
            max_word_length=20,
        )
        assert isinstance(mlt.min_word_length, int)
        assert isinstance(mlt.max_word_length, int)

    def test_mlt_numeric_validation(self) -> None:
        """MLT validates that numeric parameters are positive integers."""
        # Test min_term_freq
        with pytest.raises(ValueError, match="min_term_freq must be >= 1"):
            MoreLikeThis(id=1, min_term_freq=0)

        with pytest.raises(ValueError, match="min_term_freq must be >= 1"):
            MoreLikeThis(id=1, min_term_freq=-1)

        # Test max_query_terms
        with pytest.raises(ValueError, match="max_query_terms must be >= 1"):
            MoreLikeThis(id=1, max_query_terms=0)

        # Test min_word_length
        with pytest.raises(ValueError, match="min_word_length must be >= 1"):
            MoreLikeThis(id=1, min_word_length=0)

        # Test type validation
        with pytest.raises(TypeError, match="min_term_freq must be an integer"):
            MoreLikeThis(id=1, min_term_freq="5")  # type: ignore[arg-type]

        with pytest.raises(TypeError, match="min_term_freq must be an integer"):
            MoreLikeThis(id=1, min_term_freq=True)  # type: ignore[arg-type]

        # Valid values should work
        mlt = MoreLikeThis(
            id=1,
            min_term_freq=1,
            max_query_terms=100,
            min_word_length=1,
        )
        assert mlt.min_term_freq == 1
        assert mlt.max_query_terms == 100
        assert mlt.min_word_length == 1

    def test_mlt_custom_key_field(self) -> None:
        """MLT accepts custom key_field parameter."""
        mlt = MoreLikeThis(id=1, key_field="custom_id")
        assert mlt.key_field == "custom_id"

        # Default should be None
        mlt = MoreLikeThis(id=1)
        assert mlt.key_field is None

    def test_mlt_key_field_validation(self) -> None:
        with pytest.raises(TypeError, match="key_field must be a string"):
            MoreLikeThis(id=1, key_field=1)  # type: ignore[arg-type]

        with pytest.raises(ValueError, match="key_field cannot be empty"):
            MoreLikeThis(id=1, key_field="")
